from ..common.icb import icb
from ..common.vector import vector
from ..common.supported_features import GROUP_OF_OPCODES, FUNCT6_TO_INST_MAP


DEFAULT_CONFIGS = {
    'pc_width': 10,
    'addr_width': 12,
    'elen': 32,
    'vlen': 128,
    'xlen': 32
}


class simulator:
    def __init__(self, imem: dict = {}, dmem: dict = {}, configs: dict = DEFAULT_CONFIGS, debug_mode: bool = False) -> None:
        self.imem = imem
        self.dmem = dmem
        self.configs = configs
        self.x_reg_file = {key: 0 for key in range(32)}
        self.v_reg_file = {key: 0 for key in range(32)}
        self.pc = 0
        self.debug_mode = debug_mode


    def __apply_changes(self, changes: dict) -> None:
        if 'pc' in changes:
            self.pc = changes['pc']
        
        if 'x_reg_file' in changes:
            for rd, value in changes['x_reg_file'].items():
                self.x_reg_file[rd] = value

        if 'v_reg_file' in changes:
            for vd, value in changes['v_reg_file'].items():
                self.v_reg_file[vd] = value

        if 'dmem' in changes:
            for addr, byte in changes['dmem'].items():
                self.dmem[addr] = byte
    
    
    def run(self):
        vlen = self.configs['vlen']
        elen = self.configs['elen']
        xlen = self.configs['xlen']
        addr_width = self.configs['addr_width']
        dmem_size = 1 << addr_width
        pc_width = self.configs['pc_width']
        imem_size = 1 << pc_width

        changlog = []

        while True:
            changes = {} # Init changes dict for each instruction

            if  self.pc >= imem_size: # End of IMEM
                break
            
            if self.pc % 4 != 0: # Check if pc is aligned
                raise ValueError(f'Invalid PC: {self.pc}. Must be aligned to 4.')
            
            inst = self.imem.get(self.pc, 0) # 0 if pc is not in imem

            if inst == 0: # No more inst
                break

            opcode = icb.get_bits(inst, start=0, width=7)
            
            if opcode == 0b1010111: # v_arith
                opcode, vd, funct3, vs1_rs1_imm, vs2, vm, funct6 = simulator.__decode_inst_v_arith(inst)

                op = FUNCT6_TO_INST_MAP[funct6]
                
                vect2 = vector(self.v_reg_file[vs2], elen=elen, vlen=vlen)
                if funct3 == 0b000: # OPIVV
                    vect1 = vector(self.v_reg_file[vs1_rs1_imm], elen=elen, vlen=vlen)

                    format = 'v.v' if op == 'vmv' else 'vvm' if op == 'vmerge' else 'vv'
                    self.__debug_log(f"{op}.{format} v{vd}, {f'v{vs2}, ' if op != 'vmv' else ''}v{vs1_rs1_imm}{', v0.t' if not vm else ''}")
                    
                elif funct3 == 0b100: # OPIVX
                    vect1 = vector(vect=[icb(self.x_reg_file[vs1_rs1_imm], width=elen)] * (vlen // elen), elen=elen, vlen=vlen)

                    format = 'v.x' if op == 'vmv' else 'vxm' if op == 'vmerge' else 'vx'
                    self.__debug_log(f"{op}.{format} v{vd}, {f'v{vs2}, ' if op != 'vmv' else ''}x{vs1_rs1_imm}{', v0.t' if not vm else ''}")

                elif funct3 == 0b011: # OPIVI
                    vect1 = vector(vect=[icb(vs1_rs1_imm, width=5)] * (vlen // elen), elen=elen, vlen=vlen)

                    format = 'v.i' if op == 'vmv' else 'vim' if op == 'vmerge' else 'vi'
                    self.__debug_log(f"{op}.{format} v{vd}, {f'v{vs2}, ' if op != 'vmv' else ''}{hex(vs1_rs1_imm)}{', v0.t' if not vm else ''}")

                else:
                    raise ValueError(f'Unsupported funct3: 0b{funct3:3b}.')
                
                masks = self.__get_masks(vm)

                result = self.__vop(funct6, vect2, vect1, masks)
                
                self.__debug_log(f"Source: {f'v{vs2}':7} = {' '.join([elm.to_hex() for elm in vect2.elms])}", indent=2)
                self.__debug_log(f"Source: {f'v{vs1_rs1_imm}' if format in ['vv', 'vvm', 'v.v'] else f'x{vs1_rs1_imm}' if format in ['vx', 'vxm', 'v.x'] else 'imm5':7} = {' '.join([elm.to_hex() for elm in vect1.elms])}", indent=2)
                self.__debug_log(f"Masks : {' ':7} = {' '.join([icb(mask, elen).to_hex() for mask in masks])}", indent=2)
                self.__debug_log(f"Result: {f'v{vd}':7} = {' '.join([icb(icb.get_bits(result, elen * i, elen), elen).to_hex() for i in range(vlen // elen)])}", indent=2)

                changes['pc'] = self.pc + 4
                if self.v_reg_file[vd] != result:
                    changes['v_reg_file'] = {vd: result}

            elif opcode == 0b0000111: # v_load
                opcode, vd, width_code, rs1, lumop_rs2_vs2, vm, mop, mew, nf = simulator.__decode_inst_v_load(inst)

                masks = self.__get_masks(vm)
                
                if width_code == 0b000: # 8-bit
                    width = 8
                elif width_code == 0b101: # 16-bit
                    width = 16
                elif width_code == 0b110: # 32-bit
                    width = 32
                else:
                    raise ValueError(f'Unsupported width_code: 0b{width_code:3b}.')

                base_addr = self.x_reg_file[rs1]

                if mop == 0b00: # unit-stride
                    read_vect = self.__vload_unit_stride(vd, width, base_addr, masks)

                    self.__debug_log(f'vle{width}.v v{vd}, (x{rs1}){', v0.t' if not vm else ''}')
                    self.__debug_log(f'Source: base    = {icb(base_addr, xlen).to_hex()}', indent=2)
                    self.__debug_log(f"Masks :         = {' '.join([icb(mask, elen).to_hex() for mask in masks])}", indent=2)
                    self.__debug_log(f"Result: {f'v{vd}':7} = {' '.join([icb(icb.get_bits(read_vect, elen * i, elen), elen).to_hex() for i in range(vlen // elen)])}", indent=2)
                
                elif mop == 0b01: # indexed-unordered
                    index_vect = vector(self.v_reg_file[lumop_rs2_vs2], elen=elen, vlen=vlen)
                    read_vect = self.__vload_indexed_unordered(vd, width, base_addr, index_vect, masks)

                    self.__debug_log(f'vluxei{width}.v v{vd}, (x{rs1}), v{lumop_rs2_vs2}{', v0.t' if not vm else ''}')
                    self.__debug_log(f'Source: base    = {icb(base_addr, xlen).to_hex()}', indent=2)
                    self.__debug_log(f'Source: indexes = {' '.join([elm.to_hex() for elm in index_vect.elms])}', indent=2)
                    self.__debug_log(f"Masks :         = {' '.join([icb(mask, elen).to_hex() for mask in masks])}", indent=2)
                    self.__debug_log(f"Result: {f'v{vd}':7} = {' '.join([icb(icb.get_bits(read_vect, elen * i, elen), elen).to_hex() for i in range(vlen // elen)])}", indent=2)
                    
                elif mop == 0b10: # strided
                    stride = self.x_reg_file[lumop_rs2_vs2]
                    read_vect = self.__vload_strided(vd, width, base_addr, stride, masks)

                    self.__debug_log(f'vlse{width}.v v{vd}, (x{rs1}), x{lumop_rs2_vs2}{', v0.t' if not vm else ''}')
                    self.__debug_log(f'Source: base    = {icb(base_addr, xlen).to_hex()}', indent=2)
                    self.__debug_log(f'Source: stride  = {icb(stride, xlen).to_hex()}', indent=2)
                    self.__debug_log(f"Masks :         = {' '.join([icb(mask, elen).to_hex() for mask in masks])}", indent=2)
                    self.__debug_log(f"Result: {f'v{vd}':7} = {' '.join([icb(icb.get_bits(read_vect, elen * i, elen), elen).to_hex() for i in range(vlen // elen)])}", indent=2)
                    
                else:
                    raise ValueError(f'Unsupported mop: 0b{mop:2b}.')
                
                changes['pc'] = self.pc + 4
                if self.v_reg_file[vd] != read_vect:
                    changes['v_reg_file'] = {vd: read_vect}

            elif opcode == 0b0100111: # v_store
                opcode, vs3, width_code, rs1, sumop_rs2_vs2, vm, mop, mew, nf = simulator.__decode_inst_v_store(inst)
                
                masks = self.__get_masks(vm)
                
                if width_code == 0b000: # 8-bit
                    width = 8
                elif width_code == 0b101: # 16-bit
                    width = 16
                elif width_code == 0b110: # 32-bit
                    width = 32
                else:
                    raise ValueError(f'Unsupported width_code: 0b{width_code:3b}.')

                base_addr = self.x_reg_file[rs1]
                write_vect = vector(self.v_reg_file[vs3], elen=elen, vlen=vlen)

                if mop == 0b00: # unit-stride
                    dmem_changes = self.__vstore_unit_stride(write_vect, width, base_addr, masks)

                    self.__debug_log(f'vse{width}.v v{vs3}, (x{rs1}){', v0.t' if not vm else ''}')
                    self.__debug_log(f"Source: {f'v{vs3}':7} = {' '.join([elm.to_hex() for elm in write_vect.elms])}", indent=2)
                    self.__debug_log(f'Source: base    = {icb(base_addr, xlen).to_hex()}', indent=2)
                    self.__debug_log(f"Masks :         = {' '.join([icb(mask, elen).to_hex() for mask in masks])}", indent=2)
                    self.__debug_log(f"Result: dmem    = {', '.join([f'[{icb(addr, addr_width).to_hex()}]: {icb(byte, 8).to_hex()}' for addr, byte in dmem_changes.items()])}", indent=2)
                
                elif mop == 0b01: # indexed-unordered
                    index_vect = vector(self.v_reg_file[sumop_rs2_vs2], elen=elen, vlen=vlen)
                    dmem_changes = self.__vstore_indexed_unordered(write_vect, width, base_addr, index_vect, masks)

                    self.__debug_log(f'vsuxei{width}.v v{vs3}, (x{rs1}), v{lumop_rs2_vs2}{', v0.t' if not vm else ''}')
                    self.__debug_log(f"Source: {f'v{vs3}':7} = {' '.join([elm.to_hex() for elm in write_vect.elms])}", indent=2)
                    self.__debug_log(f'Source: base    = {icb(base_addr, xlen).to_hex()}', indent=2)
                    self.__debug_log(f'Source: indexes = {' '.join([elm.to_hex() for elm in index_vect.elms])}', indent=2)
                    self.__debug_log(f"Masks :         = {' '.join([icb(mask, elen).to_hex() for mask in masks])}", indent=2)
                    self.__debug_log(f"Result: dmem    = {', '.join([f'[{icb(addr, addr_width).to_hex()}]: {icb(byte, 8).to_hex()}' for addr, byte in dmem_changes.items()])}", indent=2)
                
                elif mop == 0b10: # strided
                    stride = self.x_reg_file[sumop_rs2_vs2]
                    dmem_changes = self.__vstore_strided(write_vect, width, base_addr, stride, masks)

                    self.__debug_log(f'vsse{width}.v v{vs3}, (x{rs1}), x{lumop_rs2_vs2}{', v0.t' if not vm else ''}')
                    self.__debug_log(f"Source: {f'v{vs3}':7} = {' '.join([elm.to_hex() for elm in write_vect.elms])}", indent=2)
                    self.__debug_log(f'Source: base    = {icb(base_addr, xlen).to_hex()}', indent=2)
                    self.__debug_log(f'Source: stride  = {icb(stride, xlen).to_hex()}', indent=2)
                    self.__debug_log(f"Result: dmem    = {', '.join([f'[{icb(addr, addr_width).to_hex()}]: {icb(byte, 8).to_hex()}' for addr, byte in dmem_changes.items()])}", indent=2)

                else:
                    raise ValueError(f'Unsupported mop: 0b{mop:2b}.')

                changes['pc'] = self.pc + 4
                changes['dmem'] = {addr: byte for addr, byte in dmem_changes.items() if self.dmem.get(addr, 0) != byte}

            elif opcode in GROUP_OF_OPCODES:
                if GROUP_OF_OPCODES[opcode] == 'u_type':
                    rd = icb.get_bits(inst, start=7, width=5)
                    imm20 = icb.get_bits(inst, start=12, width=20)

                    if opcode == 0b0110111: #lui
                        result = imm20 << 12
                        changes['pc'] = self.pc + 4
                        if self.x_reg_file[rd] != result:
                            changes['x_reg_file'] = {rd: result}

                        self.__debug_log(f'lui x{rd}, {hex(imm20)}')
                        self.__debug_log(f"Result: {f"x{rd}":7} = {icb(result, xlen).to_hex()}", indent=2)

                    elif opcode == 0b0010111: # auipc
                        raise ValueError(f'Unsupported opcode: 0b{opcode:7b} (Not implemented yet).')

                elif GROUP_OF_OPCODES[opcode] == 'i_type':
                    rd = icb.get_bits(inst, start=7, width=5)
                    funct3 = icb.get_bits(inst, start=12, width=3)
                    rs1 = icb.get_bits(inst, start=15, width=5)
                    imm12 = icb.get_bits(inst, start=20, width=12)

                    if funct3 == 0b000: # addi
                        opnd2 = icb(self.x_reg_file[rs1], xlen)
                        opnd1 = icb(imm12, 12)

                        result = (opnd2 + opnd1).repr

                        changes['pc'] = self.pc + 4
                        if self.x_reg_file[rd] != result:
                            changes['x_reg_file'] = {rd: result}

                        self.__debug_log(f'addi x{rd}, x{rs1}, {hex(imm12)}')
                        self.__debug_log(f"Source: {f"x{rs1}":7} = {opnd2.to_hex()}", indent=2)
                        self.__debug_log(f"Source: {"imm12":7} = {opnd1.__sext__(xlen).to_hex()}", indent=2)
                        self.__debug_log(f"Result: {f"x{rd}":7} = {icb(result, xlen).to_hex()}", indent=2)
                    else:
                        raise ValueError(f'Unsupported funct3: 0b{funct3:3b} (Not implemented yet).')

                else:
                    raise ValueError(f'Unsupported opcode: 0b{opcode:7b}.')

            else:
                raise ValueError(f'Unsupported opcode: 0b{opcode:7b}.')

            changlog.append(changes)

            self.__apply_changes(changes)

        return changlog
    

    def __vop(self, funct6: int, vect2: vector, vect1: vector, masks: list[int]):
        if funct6 == 0b000000: # vadd
            result_vect = vect2.__vadd__(vect1, masks)
        elif funct6 == 0b000010: # vsub
            result_vect = vect2.__vsub__(vect1, masks)
        elif funct6 == 0b000011: # vrsub
            result_vect = vect2.__vrsub__(vect1, masks)
        elif funct6 == 0b001001: # vand
            result_vect = vect2.__vand__(vect1, masks)
        elif funct6 == 0b001010: # vor
            result_vect = vect2.__vor__(vect1, masks)
        elif funct6 == 0b001011: # vxor
            result_vect = vect2.__vxor__(vect1, masks)
        elif funct6 == 0b100101: # vsll
            result_vect = vect2.__vsll__(vect1, masks)
        elif funct6 == 0b101000: # vsrl
            result_vect = vect2.__vsrl__(vect1, masks)
        elif funct6 == 0b101001: # vsra
            result_vect = vect2.__vsra__(vect1, masks)
        elif funct6 == 0b011000: # vmseq
            result_vect = vect2.__vmseq__(vect1, masks)
        elif funct6 == 0b011001: # vmsne
            result_vect = vect2.__vmsne__(vect1, masks)
        elif funct6 == 0b011010: # vmsltu
            result_vect = vect2.__vmsltu__(vect1, masks)
        elif funct6 == 0b011011: # vmslt
            result_vect = vect2.__vmslt__(vect1, masks)
        elif funct6 == 0b011100: # vmsleu
            result_vect = vect2.__vmsleu__(vect1, masks)
        elif funct6 == 0b011101: # vmsle
            result_vect = vect2.__vmsle__(vect1, masks)
        elif funct6 == 0b011110: # vmsgtu
            result_vect = vect2.__vmsgtu__(vect1, masks)
        elif funct6 == 0b011111: # vmsgt
            result_vect = vect2.__vmsgt__(vect1, masks)
        elif funct6 == 0b000100: # vminu
            result_vect = vect2.__vminu__(vect1, masks)
        elif funct6 == 0b000101: # vmin
            result_vect = vect2.__vmin__(vect1, masks)
        elif funct6 == 0b000110: # vmaxu
            result_vect = vect2.__vmaxu__(vect1, masks)
        elif funct6 == 0b000111: # vmax
            result_vect = vect2.__vmax__(vect1, masks)
        elif funct6 == 0b010111: # vmerge
            result_vect = vect2.__vmerge__(vect1, masks)
        elif funct6 == 0b010111: # vmv
            result_vect = vect1
        else:
            raise ValueError(f'Unsupported funct6: 0b{funct6:6b}.')
        
        return result_vect.to_register()

    
    def __vload_unit_stride(self, vd: int, width: int, base_addr: int, masks: list[int]) -> int:
        dmem_size = 1 << self.configs['addr_width']
        vlen = self.configs['vlen']
        elen = self.configs['elen']
        num_of_elms = vlen // elen

        read_vect = self.v_reg_file[vd]

        for i in range(num_of_elms):
            addr = (base_addr + i * (width // 8)) % dmem_size # Ignore higher address bits
            # if addr >= dmem_size: # TODO: Uncomment after constrain dmem access range
            #     raise ValueError(f"The address: {addr} is out of DMEM (DMEM size is {dmem_size})")

            if masks[i] == 1:
                elm_i = 0
                for j in range(width // 8):
                    elm_i |= self.dmem.get(addr + j, 0) << (j * 8)
            
                clear_mask = ~((1 << elen) - 1 << (i * elen))
                read_vect &= clear_mask # Clear old element before override
                read_vect |= icb(elm_i, width).__sext__(elen).repr << (i * elen)

        return read_vect
    

    def __vload_indexed_unordered(self, vd: int, width: int, base_addr: int, index_vect: vector, masks: list[int]) -> int:
        dmem_size = 1 << self.configs['addr_width']
        vlen = self.configs['vlen']
        elen = self.configs['elen']
        num_of_elms = vlen // elen

        read_vect = self.v_reg_file[vd]

        for i in range(num_of_elms):
            addr = (base_addr + index_vect.get_element(i).repr) % dmem_size # Ignore higher address bits
            # if addr >= dmem_size: # TODO: Uncomment after constrain dmem access range
            #     raise ValueError(f"The address: {addr} is out of DMEM (DMEM size is {dmem_size})")

            if masks[i] == 1:
                elm_i = 0
                for j in range(width // 8):
                    elm_i |= self.dmem.get(addr + j, 0) << (j * 8)
            
                clear_mask = ~((1 << elen) - 1 << (i * elen))
                read_vect &= clear_mask # Clear old element before override
                read_vect |= icb(elm_i, width).__sext__(elen).repr << (i * elen)

        return read_vect
    

    def __vload_strided(self, vd: int, width: int, base_addr: int, stride: int, masks: list[int]) -> int:
        dmem_size = 1 << self.configs['addr_width']
        vlen = self.configs['vlen']
        elen = self.configs['elen']
        num_of_elms = vlen // elen

        read_vect = self.v_reg_file[vd]

        for i in range(num_of_elms):
            addr = (base_addr + i * stride) % dmem_size # Ignore higher address bits
            # if addr >= dmem_size: # TODO: Uncomment after constrain dmem access range
            #     raise ValueError(f"The address: {addr} is out of DMEM (DMEM size is {dmem_size})")

            if masks[i] == 1:
                elm_i = 0
                for j in range(width // 8):
                    elm_i |= self.dmem.get(addr + j, 0) << (j * 8)
            
                clear_mask = ~((1 << elen) - 1 << (i * elen))
                read_vect &= clear_mask # Clear old element before override
                read_vect |= icb(elm_i, width).__sext__(elen).repr << (i * elen)

        return read_vect
    

    def __vstore_unit_stride(self, write_vect: vector, width: int, base_addr: int, masks: list[int]) -> dict:
        dmem_size = 1 << self.configs['addr_width']
        vlen = self.configs['vlen']
        elen = self.configs['elen']
        num_of_elms = vlen // elen

        dmem_changes = {}
        for i in range(num_of_elms):
            addr = (base_addr + i * (width // 8)) % dmem_size # Ignore higher address bits
            # if addr >= dmem_size: # TODO: Uncomment after constrain dmem access range
            #     raise ValueError(f"The address: {addr} is out of DMEM (DMEM size is {dmem_size})")

            if masks[i]:
                write_elm = write_vect.get_element(i).repr
                for j in range(width // 8):
                    byte = icb.get_bits(write_elm, start=8 * j, width=8)
                    dmem_changes[addr + j] = byte

        return dmem_changes
    

    def __vstore_indexed_unordered(self, write_vect: vector, width: int, base_addr: int, index_vect: vector, masks: list[int]) -> dict:
        dmem_size = 1 << self.configs['addr_width']
        vlen = self.configs['vlen']
        elen = self.configs['elen']
        num_of_elms = vlen // elen

        dmem_changes = {}
        for i in range(num_of_elms):
            addr = (base_addr + index_vect.get_element(i).repr) % dmem_size # Ignore higher address bits
            # if addr >= dmem_size: # TODO: Uncomment after constrain dmem access range
            #     raise ValueError(f"The address: {addr} is out of DMEM (DMEM size is {dmem_size})")

            if masks[i] == 1:
                write_elm = write_vect.get_element(i).repr
                for j in range(width // 8):
                    byte = icb.get_bits(write_elm, start=8 * j, width=8)
                    dmem_changes[addr + j] = byte

        return dmem_changes
    

    def __vstore_strided(self, write_vect: vector, width: int, base_addr: int, stride: int, masks: list[int]) -> dict:
        dmem_size = 1 << self.configs['addr_width']
        vlen = self.configs['vlen']
        elen = self.configs['elen']
        num_of_elms = vlen // elen

        dmem_changes = {}
        for i in range(num_of_elms):
            addr = (base_addr + i * stride) % dmem_size # Ignore higher address bits
            # if addr >= dmem_size: # TODO: Uncomment after constrain dmem access range
            #     raise ValueError(f"The address: {addr} is out of DMEM (DMEM size is {dmem_size})")

            if masks[i] == 1:
                write_elm = write_vect.get_element(i).repr
                for j in range(width // 8):
                    byte = icb.get_bits(write_elm, start=8 * j, width=8)
                    dmem_changes[addr + j] = byte

        return dmem_changes


    def __get_masks(self, vm: int) -> list[int]:
        vlen = self.configs['vlen']
        elen = self.configs['elen']

        if vm:
            return [1] * (vlen // elen)
        
        return [
            icb.get_bits(self.v_reg_file[0], start=(i * elen), width=1)
            for i in range(vlen // elen)
        ]
    
    
    def __debug_log(self, msg: str, indent: int = 0) -> None:
        if self.debug_mode:
            print(' ' * indent + msg)


    @staticmethod
    def __decode_inst_v_arith(inst: int) -> tuple:
        opcode = icb.get_bits(inst, start=0, width=7)
        vd = icb.get_bits(inst, start=7, width=5)
        funct3 = icb.get_bits(inst, start=12, width=3)
        vs1_rs1_imm = icb.get_bits(inst, start=15, width=5)
        vs2 = icb.get_bits(inst, start=20, width=5)
        vm = icb.get_bits(inst, start=25, width=1)
        funct6 = icb.get_bits(inst, start=26, width=6)

        return opcode, vd, funct3, vs1_rs1_imm, vs2, vm, funct6


    @staticmethod
    def __decode_inst_v_load(inst: int) -> tuple:
        opcode = icb.get_bits(inst, start=0,width=7)
        vd = icb.get_bits(inst, start=7,width=5)
        width_code = icb.get_bits(inst, start=12,width=3)
        rs1 = icb.get_bits(inst, start=15,width=5)
        lumop_rs2_vs2 = icb.get_bits(inst, start=20,width=5)
        vm = icb.get_bits(inst, start=25,width=1)
        mop = icb.get_bits(inst, start=26,width=2)
        mew = icb.get_bits(inst, start=28,width=1)
        nf = icb.get_bits(inst, start=29,width=3)

        return opcode, vd, width_code, rs1, lumop_rs2_vs2, vm, mop, mew, nf


    @staticmethod
    def __decode_inst_v_store(inst: int) -> tuple:
        opcode = icb.get_bits(inst, start=0, width=7)
        vs3 = icb.get_bits(inst, start=7, width=5)
        width_code = icb.get_bits(inst, start=12, width=3)
        rs1 = icb.get_bits(inst, start=15, width=5)
        sumop_rs2_vs2 = icb.get_bits(inst, start=20, width=5)
        vm = icb.get_bits(inst, start=25, width=1)
        mop = icb.get_bits(inst, start=26, width=2)
        mew = icb.get_bits(inst, start=28, width=1)
        nf = icb.get_bits(inst, start=29, width=3)

        return opcode, vs3, width_code, rs1, sumop_rs2_vs2, vm, mop, mew, nf
    
    


from ..common.supported_features import SUPPORTED_INSTS, MOPS, WIDTH_CODES


def translate_data(data):
    dmem = ''
    for label, data_def in data.items():
        for item in data_def['items']:
            for _ in range(data_def['width'] // 8):
                dmem += bin(int(item % (2 ** 8)))[2:].zfill(8) + '\n'
                item //= (2 ** 8)

    return dmem


def icb_of_v_arith(format, funct6, vd, vs2, vs1, rs1, imm, vm):
    if format in ['vv', 'vvm', 'v.v']:
        funct3 = 0b000
        bit_19_15 = vs1
    elif format in ['vx', 'vxm', 'v.x']:
        funct3 = 0b100
        bit_19_15 = rs1
    elif format in ['vi', 'vim', 'v.i']:
        funct3 = 0b011
        bit_19_15 = imm

    icb = 0b1010111 # opcode
    icb |= (vd << 7)
    icb |= (funct3 << 12)
    icb |= (bit_19_15 << 15)
    icb |= (vs2 << 20)
    icb |= (int(vm) << 25)
    icb |= (funct6 << 26)

    return icb


def icb_of_v_store(mop, width, vs3, rs1, rs2, vs2, vm, sumop=0b00000, nf=0b000, mew=0b0):
    if mop == 0b00:
        bit_24_20 = sumop
    elif mop == 0b01:
        bit_24_20 = vs2
    elif mop == 0b10:
        bit_24_20 = rs2

    icb = 0b0100111 # opcode
    icb |= (vs3 << 7)
    icb |= (width << 12)
    icb |= (rs1 << 15)
    icb |= (bit_24_20 << 20)
    icb |= (int(vm) << 25)
    icb |= (mop << 26)
    icb |= (mew << 28)
    icb |= (nf << 29)

    return icb


def icb_of_v_load(mop, width, vd, rs1, rs2, vs2, vm, lumop=0b00000, nf=0b00, mew=0b0):
    if mop == 0b00:
        bit_24_20 = lumop
    elif mop == 0b01:
        bit_24_20 = vs2
    elif mop == 0b10:
        bit_24_20 = rs2

    print(f'{width:b=}')

    icb = 0b0000111 # opcode
    icb |= (vd << 7)
    icb |= (width << 12)
    icb |= (rs1 << 15)
    icb |= (bit_24_20 << 20)
    icb |= (int(vm) << 25)
    icb |= (mop << 26)
    icb |= (mew << 28)
    icb |= (nf << 29)

    return icb


def icb_of_u_type(opcode, rd, imm20):
    icb = opcode
    icb |= (rd << 7)
    icb |= (imm20 << 12)

    return icb


def icb_of_i_type(opcode, funct3, rd, rs1, imm12):
    icb = opcode
    icb |= (rd << 7)
    icb |= (funct3 << 12)
    icb |= (rs1 << 15)
    icb |= (imm12 << 20)

    return icb


def translate_text(text):
    imem = ''
    for inst in text:
        print(inst)

        inst_params = SUPPORTED_INSTS[inst['op']]
        group = inst_params['group']
        if group.startswith('v_arith'):
            funct6 = inst_params['funct6']
            vs2 = inst['vs2'] if 'vs2' in inst else 0
            vs1 = inst['vs1'] if 'vs1' in inst else None
            rs1 = inst['rs1'] if 'rs1' in inst else None
            imm = inst['imm'] if 'imm' in inst else None
            vm = inst['vm'] if 'vm' in inst else True
            icb = icb_of_v_arith(inst['format'], funct6, inst['vd'], vs2, vs1, rs1, imm, vm)
            
        elif group.startswith('v_store'):
            mop = MOPS[inst['op']]
            rs2 = inst['rs2'] if 'rs2' in inst else None
            vs2 = inst['vs2'] if 'vs2' in inst else None
            icb = icb_of_v_store(mop, WIDTH_CODES[inst['width']], inst['vs3'], inst['rs1'], rs2, vs2, inst['vm'])

        elif group.startswith('v_load'):
            mop = MOPS[inst['op']]
            rs2 = inst['rs2'] if 'rs2' in inst else None
            vs2 = inst['vs2'] if 'vs2' in inst else None
            icb = icb_of_v_load(mop, WIDTH_CODES[inst['width']], inst['vd'], inst['rs1'], rs2, vs2, inst['vm'])
        
        # Below cases for scalar instructions
        elif group.startswith('u_type'):
            icb = icb_of_u_type(inst_params['opcode'], inst['rd'], inst['imm20'])

        elif group.startswith('i_type'):
            icb = icb_of_i_type(inst_params['opcode'], inst_params['funct3'], inst['rd'], inst['rs1'], inst['imm12'])
        
        imem += bin(icb)[2:].zfill(32) + '\n'

    return imem


from .confparser import parse_config_file
from .vrtg import generate_operator_sequence, add_operands, insert_vm_change_placeholders, replace_vm_change_placeholders, display
from .init_data_generator import trace_uninitialzed_operands, init_data


def generate(configs):
    operator_seq = generate_operator_sequence(configs)
    main_text_seq = add_operands(operator_seq, configs)
    main_text_seq = insert_vm_change_placeholders(main_text_seq, configs)
    vregs2init, xreg2init, vms2init = trace_uninitialzed_operands(main_text_seq)
    data_section, init_text_seq = init_data(vregs2init, xreg2init, vms2init, configs)
    main_text_seq = replace_vm_change_placeholders(main_text_seq)
    
    return data_section, init_text_seq, main_text_seq




from random import random, randint, choice, choices
import re


def ask_user(prompt: str):
    res = input(f'{prompt} [Y/n] ').strip().lower()
    if res in ['y', 'yes', '']:
        return True
    elif res in ['n', 'no']:
        return False
    else:
        print('Invalid option.')
        return ask_user(prompt)
    

def repeated_random(min, max, used_set: set, repeat_rate):
    r = random()
    if r <= repeat_rate and len(used_set) > 0:
        return choice(list(used_set))
    else:
        return randint(min, max)
    
    
def random_imm(width, data_variant_weights): # TODO: Test sign of result
    if width == 8:
        variant = choices(['zero', 'neg', 'pos'], data_variant_weights[:-1])[0]
        if variant == 'zero':
            return 0, False
        elif variant == 'neg':
            return 2 ** 7 + randint(0, 2 ** 7 - 1), False
        else:
            return randint(0, 2 ** 7 - 1), False
    
    elif width == 16:
        variant = choices(['zero', 'neg', 'pos', 'big_num'], data_variant_weights)[0]
        if variant == 'zero':
            return 0, False
        elif variant == 'neg':
            return 2 ** 11 + randint(0, 2 ** 11 - 1), True
        elif variant == 'pos':
            return randint(0, 2 ** 11 - 1), False
        else:
            return randint(2 ** 12, 2 ** 16 - 1), True

    elif width == 32:
        variant = choices(['zero', 'neg', 'pos', 'big_num'], data_variant_weights)[0]
        if variant == 'zero':
            return 0, False
        elif variant == 'neg':
            return 2 ** 11 + randint(0, 2 ** 11 - 1), True
        elif variant == 'pos':
            return randint(0, 2 ** 11 - 1), False
        else:
            return randint(2 ** 12, 2 ** 32 - 1), True

    else:
        pass
        # TODO: Raise error here


def split_imm(imm):
    hi20b = imm // (2 ** 12)
    lo12b = imm % (2 ** 12)
    if lo12b >= 2 ** 11:
        hi20b += 1

    return hi20b, lo12b


def sign_mag_2_two_comp(value, width):
    if value > 0:
        two_comp = value
    else:
        two_comp = (~abs(value) | (1 << (width - 1))) + 1

    return two_comp % (2 ** width)


def code_number(number_as_str: str, width):
    '''
    Convert number as string into int coded (two-comp) binary.
    Return a non-negative int
    '''
    number_match = re.match(r'(?P<bin>0b[01]+)|(?P<hex>0x[0-9a-fA-F]+)|(?P<dec>-?\d+)', number_as_str)

    if not number_match:
        return -1 # Error code

    if number_match.group('bin'):
        return int(number_match.group('bin'), 2)
    elif number_match.group('hex'):
        return int(number_match.group('hex'), 16)
    elif number_match.group('dec'):
        return sign_mag_2_two_comp(int(number_match.group('dec')), width)


def print_error_and_exit(line_number, statement, error_descs):
    print(f'SyntaxError at {line_number}: {statement}')
    for desc in error_descs:
        print(desc)
    exit(1)


import re

from ..common.supported_features import DATA_TYPES, TYPE_WIDTHS, SUPPORTED_INSTS, VECTOR_INSTS, SCALAR_INSTS
from ..common.helpers import code_number, print_error_and_exit


def sectionify(asm: str):
    text_section = {}
    data_section = {}
    current_section = 'text' # Default is .text section
    
    lines = asm.splitlines()
    for i in range(len(lines)):
        line = lines[i].strip()
        
        if line == '':
            continue

        section_match = re.match(r'^\.(?P<section>text|data)\b', line)
        if section_match:
            current_section = section_match.group('section') # Update current section
        else: # If in a section, add the line to it
            if current_section == 'text':
                text_section[i+1] = line
            elif current_section == 'data':
                data_section[i+1] = line
    
    return text_section, data_section


def parse_data(data_section):
    current_addr = 0
    data = {}
    for line_number, statement in data_section.items():
        # TODO: like .text section, label can be missing; label and data definition can be on multi lines
        statement_match = re.match(r'(?P<label>\w+):\s*(?P<type>\.\w+)\s*(?P<data>[^#]*)(?:\s*#(?P<comment>.*))?', statement)
        
        # Check overall syntax
        if not statement_match:
            print_error_and_exit(line_number, statement, [
                'Right syntax: <label>: <type> <data_item_1>[, <data_item_2>, ...,<data_item_n>][# <comment>]'
                '<label> must start with a letter or _ not a number'
            ])
        
        # Check supported types
        type = statement_match.group('type')
        if type not in DATA_TYPES:
            print_error_and_exit(line_number, statement, [
                f"Unsupported type: '{type}'"
            ])

        # Check data sequence syntax
        data_seq = statement_match.group('data')
        data_match = re.match(r'(?:\s*(?:0b[01]+|0x[0-9a-fA-F]+|\-?\d+)\s*,\s*)*(?:0b[01]+|0x[0-9a-fA-F]+|\-?\d+)\s*', data_seq)

        if not data_match:
            print_error_and_exit(line_number, statement, [
                f"Some thing wrong here: '{type}'"
            ]) # TODO: Modify error message
        
        items = []
        for item in data_seq.split(','):
            item = item.strip()

            number = code_number(item, TYPE_WIDTHS[type])
            if number == -1:
                print_error_and_exit(line_number, statement, [
                    f"Invalid number format '{item}'"
                ])

            # TODO: check overflow

            items.append(number)

        data[statement_match.group('label')] = {
            'addr': current_addr,
            'width': TYPE_WIDTHS[type],
            'items': items
        }

        current_addr += len(items) * TYPE_WIDTHS[type] // 8

    return data


def syntax_2_regex(syntax: str, dynamic_params):
    syntax = syntax.replace('.', r'\.')
    if 'formats' in dynamic_params:
        syntax = syntax.replace('<format>', f"(?P<format>{'|'.join(format.replace('.', '\\.') for format in dynamic_params['formats'])})")
    if 'widths' in dynamic_params:
        syntax = syntax.replace('<width>', f"(?P<width>{'|'.join([str(width) for width in dynamic_params['widths']])})")
    syntax = syntax.replace(' <vd>', r'\s+v(?P<vd>\d+)')
    syntax = syntax.replace(' <rd>', r'\s+x(?P<rd>\d+)')
    syntax = syntax.replace(' <vs3>', r'\s+v(?P<vs3>\d+)')
    syntax = syntax.replace('<vs2>', r'v(?P<vs2>\d+)')
    syntax = syntax.replace('<rs2>', r'x(?P<rs2>\d+)')
    syntax = syntax.replace('<rs1>', r'x(?P<rs1>\d+)')
    syntax = syntax.replace(r'(x(?P<rs1>\d+))', r'(?:\(x(?P<rs1>\d+)\))')
    syntax = syntax.replace('<imm20>', r'(?P<imm20>0b[01]+|0x[0-9a-fA-F]+|\-?\d+)')
    syntax = syntax.replace('<imm12>', r'(?P<imm12>0b[01]+|0x[0-9a-fA-F]+|\-?\d+)')
    syntax = syntax.replace('<vs1|rs1|imm>', r'(?:v(?P<vs1>\d+)|x(?P<rs1>\d+)|(?P<imm>0b[01]+|0x[0-9a-fA-F]+|\-?\d+))')
    syntax = syntax.replace('<vs1|rs1>', r'(?:v(?P<vs1>\d+)|x(?P<rs1>\d+))')
    syntax = syntax.replace('<rs1|imm>', r'(?:x(?P<rs1>\d+)|(?P<imm>0b[01]+|0x[0-9a-fA-F]+|-?\d+))')
    syntax = syntax.replace(r', v0\.t', r'(?:\s*,\s*(?P<vm>v0\.t))')
    syntax = syntax.replace(r'[(?:\s*,\s*(?P<vm>v0\.t))]', r'(?:\s*,\s*(?P<vm>v0\.t))?')
    syntax = syntax.replace(',', r'\s*,')
    syntax = syntax.replace(' ', r'\s*')
    syntax += r'(?:\s*#(?P<comment>.*))?'

    return syntax


def replace_relocation_functions(inst, data):
    hi_reloc_funct_match = re.match(r'.*%hi\((?P<symbol>\w+)\)', inst)
    if hi_reloc_funct_match:
        symbol = hi_reloc_funct_match.group('symbol')
        inst = re.sub(r'%hi\((?P<symbol>\w+)\)', str(data[symbol]['addr'] >> 12), inst)
    
    lo_reloc_funct_match = re.match(r'.*%lo\((?P<symbol>\w+)\)', inst)
    if lo_reloc_funct_match:
        symbol = lo_reloc_funct_match.group('symbol')
        inst = re.sub(r'%lo\((?P<symbol>\w+)\)', str(data[symbol]['addr'] % (2 ** 12)), inst)

    return inst  


def parse_text(text_section, data):
    inst_seq = {}
    labels = {}
    pc_map_line = {}
    pc = 0
    
    for line_number, statement in text_section.items():
        # Check overall syntax: [<label>:][<inst>]
        syntax_match = re.match(r'(?:(?P<label>\w+)\s*:\s*)?(?:(?P<inst>.*))?', statement)
        if not syntax_match:
            print_error_and_exit(line_number, statement, [
                'Right syntax: [<label>: ][<inst>]'
            ])

        label = syntax_match.group('label')
        inst = syntax_match.group('inst')

        if label:
            if label in data:
                print_error_and_exit(line_number, statement, [
                    f"Duplicated label: '{label}'"
                ])
            
            labels[label] = pc

        if inst:
            inst_seq[pc] = inst
            pc_map_line[pc] = line_number
            pc += 4

    text = []
    for addr, inst in inst_seq.items():
        # Check supported operator
        op_match = re.match(r'^[a-z]+', inst)
        op = op_match.group(0)
        if op in VECTOR_INSTS:
            inst_params = VECTOR_INSTS[op]
            inst_pattern = syntax_2_regex(inst_params['syntax'], inst_params)

            # Check syntax
            inst_match = re.match(inst_pattern, inst)
            if not inst_match:
                print_error_and_exit(pc_map_line[addr], text_section[pc_map_line[addr]], [
                    f"The instruction must follow this pattern: {inst_params['syntax']}[# <comment>]"
                ])

            # Check valid params
            cuptured_params = inst_match.groupdict()
            print(inst, op, cuptured_params)
            inst_params = {'op': op}
            if 'format' in cuptured_params and not cuptured_params['format'] == None:
                inst_params['format'] = cuptured_params['format']
            
            if 'width' in cuptured_params and not cuptured_params['width'] == None:
                inst_params['width'] = int(cuptured_params['width'])

            if op != 'vmv':
                inst_params['vm'] = (cuptured_params['vm'] == None)
            
            if 'vd' in cuptured_params and not cuptured_params['vd'] == None:
                vd = int(cuptured_params['vd'])
                if vd < 0 or vd > 31:
                    print_error_and_exit(pc_map_line[addr], text_section[pc_map_line[addr]], [
                        f'Invalid vd: v{vd}'
                    ])
                inst_params['vd'] = vd
                
            if 'vs3' in cuptured_params and not cuptured_params['vs3'] == None:
                vs3 = int(cuptured_params['vs3'])
                if vs3 < 0 or vs3 > 31:
                    print_error_and_exit(pc_map_line[addr], text_section[pc_map_line[addr]], [
                        f'Invalid vs3: v{vs3}'
                    ])
                inst_params['vs3'] = vs3

            if 'vs2' in cuptured_params and not cuptured_params['vs2'] == None:
                vs2 = int(cuptured_params['vs2'])
                if vs2 < 0 or vs2 > 31:
                    print_error_and_exit(pc_map_line[addr], text_section[pc_map_line[addr]], [
                        f'Invalid vs2: v{vs2}'
                    ])
                inst_params['vs2'] = vs2

            if 'vs1' in cuptured_params and not cuptured_params['vs1'] == None:
                vs1 = int(cuptured_params['vs1'])
                if vs1 < 0 or vs1 > 31:
                    print_error_and_exit(pc_map_line[addr], text_section[pc_map_line[addr]], [
                        f'Invalid vs1: v{vs1}'
                    ])
                inst_params['vs1'] = vs1
                
            if 'rs2' in cuptured_params and not cuptured_params['rs2'] == None:
                rs2 = int(cuptured_params['rs2'])
                if rs2 < 0 or rs2 > 31:
                    print_error_and_exit(pc_map_line[addr], text_section[pc_map_line[addr]], [
                        f'Invalid rs2: x{rs2}'
                    ])
                inst_params['rs2'] = rs2
                
            if 'rs1' in cuptured_params and not cuptured_params['rs1'] == None:
                rs1 = int(cuptured_params['rs1'])
                if rs1 < 0 or rs1 > 31:
                    print_error_and_exit(pc_map_line[addr], text_section[pc_map_line[addr]], [
                        f'Invalid rs1: x{rs1}'
                    ])
                inst_params['rs1'] = rs1

            if 'imm' in cuptured_params and not cuptured_params['imm'] == None:
                imm = code_number(cuptured_params['imm'], 5)
                if imm == -1 or imm > 31:
                    print_error_and_exit(pc_map_line[addr], text_section[pc_map_line[addr]], [
                        'Invalid imm'
                    ])
                inst_params['imm'] = imm

        elif op in SCALAR_INSTS:
            # Replace Relocation Functions by Addresses
            inst = replace_relocation_functions(inst, data)
            
            inst_params = SCALAR_INSTS[op]
            inst_pattern = syntax_2_regex(inst_params['syntax'], inst_params)

            # Check syntax
            inst_match = re.match(inst_pattern, inst)
            if not inst_match:
                print_error_and_exit(pc_map_line[addr], text_section[pc_map_line[addr]], [
                    f"The instruction must follow this pattern: {inst_params['syntax']}[# <comment>]"
                ])

            # Check valid params
            cuptured_params = inst_match.groupdict()
            inst_params = {'op': op}
            
            if 'rd' in cuptured_params and not cuptured_params['rd'] == None:
                rd = int(cuptured_params['rd'])
                if rd < 0 or rd > 31:
                    print_error_and_exit(pc_map_line[addr], text_section[pc_map_line[addr]], [
                        f'Invalid rd: x{rd}'
                    ])
                inst_params['rd'] = rd

            if 'rs2' in cuptured_params and not cuptured_params['rs2'] == None:
                rs2 = int(cuptured_params['rs2'])
                if rs2 < 0 or rs2 > 31:
                    print_error_and_exit(pc_map_line[addr], text_section[pc_map_line[addr]], [
                        f'Invalid rs2: x{rs2}'
                    ])
                inst_params['rs2'] = rs2

            if 'rs1' in cuptured_params and not cuptured_params['rs1'] == None:
                rs1 = int(cuptured_params['rs1'])
                if rs1 < 0 or rs1 > 31:
                    print_error_and_exit(pc_map_line[addr], text_section[pc_map_line[addr]], [
                        f'Invalid rs1: x{rs1}'
                    ])
                inst_params['rs1'] = rs1

            if 'imm20' in cuptured_params and not cuptured_params['imm20'] == None:
                imm20 = code_number(cuptured_params['imm20'], 20)
                if imm20 == -1 or imm20 > 2 ** 20 - 1:
                    print_error_and_exit(pc_map_line[addr], text_section[pc_map_line[addr]], [
                        'Invalid imm20'
                    ])
                inst_params['imm20'] = imm20

            if 'imm12' in cuptured_params and not cuptured_params['imm12'] == None:
                imm12 = code_number(cuptured_params['imm12'], 12)
                if imm12 == -1 or imm12 > 2 ** 12 - 1:
                    print_error_and_exit(pc_map_line[addr], text_section[pc_map_line[addr]], [
                        'Invalid imm12'
                    ])
                inst_params['imm12'] = imm12

            if 'imm5' in cuptured_params and not cuptured_params['imm5'] == None:
                imm5 = code_number(cuptured_params['imm5'], 5)
                if imm5 == -1 or imm5 > 2 ** 5 - 1:
                    print_error_and_exit(pc_map_line[addr], text_section[pc_map_line[addr]], [
                        'Invalid imm5'
                    ])
                inst_params['imm5'] = imm5

        else:
            print_error_and_exit(pc_map_line[addr], text_section[pc_map_line[addr]], [
                f"Unsupported operator: '{op}'"
            ])

        text.append(inst_params)
    
    return text

        


from random import choices, randint
from ..common.supported_features import DATA_TYPES, TYPE_WIDTHS
from ..common.helpers import random_imm, split_imm

def trace_uninitialzed_operands(text_seq):
    vregs2init = set()
    xregs2init = set()
    vms2init = []

    for inst in reversed(text_seq):
        if inst['operator'] == 'placeholder.chgvm':
            vms2init.insert(0, inst['vm_label'])
            continue

        if 'vs3' in inst:
            vregs2init.add(inst['vs3'])
        if 'vs2' in inst:
            vregs2init.add(inst['vs2'])
        if 'vs1' in inst:
            vregs2init.add(inst['vs1'])
        if 'vd' in inst and inst['vd'] in vregs2init:
            vregs2init.remove(inst['vd'])

        if 'rs2' in inst:
            xregs2init.add(inst['rs2'])
        if 'rs1' in inst:
            xregs2init.add(inst['rs1'])
        if 'rd' in inst and inst['rd'] in vregs2init:
            xregs2init.remove(inst['rd'])

    return vregs2init, xregs2init, vms2init


# Currently, init values ignore dmem_access_range
def init_data(vregs2init, xregs2init, vms2init, configs):
    data_section = {}
    text_section = []

    addr = 0

    print(DATA_TYPES, configs['data_type_weights'])
    
    for vreg in vregs2init:
        data_type = choices(DATA_TYPES, configs['data_type_weights'])[0]
        data_items = []
        for _ in range(4):
            item, __ = random_imm(TYPE_WIDTHS[data_type], configs['data_variant_weights'])
            data_items.append(item)

        data_def = {
            'addr': addr,
            'type': data_type,
            'items': data_items,
        }
        data_section[f'_v{vreg}_'] = data_def

        imm20 = f"%hi(_v{vreg}_)"
        imm12 = f"%lo(_v{vreg}_)"

        text_section.extend([
            {'operator': 'lui', 'rd': 1, 'imm20': imm20},
            {'operator': 'addi', 'rd': 1, 'rs1': 1, 'imm12': imm12},
            {'operator': 'vle', 'width' : TYPE_WIDTHS[data_type], 'vd': vreg, 'rs1': 1, 'vm': True}
        ])

        addr += TYPE_WIDTHS[data_type] * len(data_items)

    for xreg in xregs2init:
        if xreg == 0:
            continue
        
        data_type = choices(DATA_TYPES, configs['data_type_weights'])[0]
        value, is_big = random_imm(TYPE_WIDTHS[data_type], configs['data_variant_weights'])
        if is_big:
            hi20b, lo12b = split_imm(value)

            text_section.extend([
                {'operator': 'lui', 'rd': xreg, 'imm20': hi20b},
                {'operator': 'addi', 'rd': xreg, 'rs1': xreg, 'imm12': lo12b},
            ])
        else:
            text_section.append(
                {'operator': 'addi', 'rd': xreg, 'rs1': 0, 'imm12': value}
            )

    for vm_label in vms2init:
        data_def = {
            'addr': addr,
            'type': '.byte',
            'items': [randint(0, 1 << 8) for _ in range(4)]
        }
        data_section[vm_label] = data_def
        addr += TYPE_WIDTHS[data_type] * len(data_items)

    print(data_section)

    return data_section, text_section


import re

DEFAULT_WIDTH = 32 # bits


class icb: # int-coded-binary
    def __init__(self, value: int | str = 0, width: int = DEFAULT_WIDTH):
        if isinstance(value, str):
            temp = icb.parse_from_str(str)
            self.repr = icb._sign_mag_to_two_comp(temp, width)
        elif isinstance(value, int):
            self.repr = icb._sign_mag_to_two_comp(value, width)
        self.width = width

    
    # zero-extend
    def __zext__(self, new_width: int) -> 'icb':
        if self.width > new_width:
            raise ValueError(f"New width ({new_width}) must be greater than current width ({self.width})")
        
        return icb(self.repr, new_width)


    # sign-extend
    def __sext__(self, new_width: int) -> 'icb':
        if self.width > new_width:
            raise ValueError(f"New width ({new_width}) must be greater than current width ({self.width})")
        
        if self.width == new_width:
            return icb(self.repr, self.width)
        
        sign_bit = self.repr >> (self.width - 1) 
        if sign_bit:
            new_repr = (self.repr | (~0 << self.width)) % (1 << new_width)
        else:
            new_repr = self.repr
        
        return icb(new_repr, new_width)

    
    def __add__(self, other: 'icb') -> 'icb':
        if self.width < other.width:
            raise ValueError(f"The width of the first operand ({self.width}) must be greater than or equal the width of the second operand ({other.width}).")
        
        ext_other = other.__sext__(self.width)
        
        return icb(self.repr + ext_other.repr, self.width)
    

    def __sub__(self, other: 'icb') -> 'icb':
        if self.width < other.width:
            raise ValueError(f"The width of the first operand ({self.width}) must be greater than or equal the width of the second operand ({other.width}).")
        
        ext_other = other.__sext__(self.width)
        
        return icb(self.repr - ext_other.repr, self.width)
    

    def __and__(self, other: 'icb') -> 'icb':
        if self.width < other.width:
            raise ValueError(f"The width of the first operand ({self.width}) must be greater than or equal the width of the second operand ({other.width}).")
        
        ext_other = other.__sext__(self.width)
        
        return icb(self.repr & ext_other.repr, self.width)
    

    def __or__(self, other: 'icb') -> 'icb':
        if self.width < other.width:
            raise ValueError(f"The width of the first operand ({self.width}) must be greater than or equal the width of the second operand ({other.width}).")
        
        ext_other = other.__sext__(self.width)
        
        return icb(self.repr | ext_other.repr, self.width)
    

    def __xor__(self, other: 'icb') -> 'icb':
        if self.width < other.width:
            raise ValueError(f"The width of the first operand ({self.width}) must be greater than or equal the width of the second operand ({other.width}).")
        
        ext_other = other.__sext__(self.width)
        
        return icb(self.repr ^ ext_other.repr, self.width)
    

    def __sll__(self, other: 'icb') -> 'icb':
        ext_other = other.__zext__(self.width)
        
        return icb(self.repr << (ext_other.repr & 31), self.width)
    

    def __srl__(self, other: 'icb') -> 'icb':
        ext_other = other.__zext__(self.width)
        
        return icb(self.repr >> (ext_other.repr & 31), self.width)
    

    def __sra__(self, other: 'icb') -> 'icb':

        sign = (self.repr >> (self.width - 1)) & 1
        shamt = other.repr & 31

        sign_ext_part = ((sign << shamt) - 1) << (self.width - shamt)
        
        temp = self.__srl__(other)

        return icb(temp.repr | sign_ext_part, temp.width) 
    

    def __seq__(self, other: 'icb') -> 'icb':
        if self.width < other.width:
            raise ValueError(f"The width of the first operand ({self.width}) must be greater than or equal the width of the second operand ({other.width}).")
        
        ext_other = other.__sext__(self.width)
        
        return icb(1 if self.repr == ext_other.repr else 0, self.width)
    

    def __sne__(self, other: 'icb') -> 'icb':
        if self.width < other.width:
            raise ValueError(f"The width of the first operand ({self.width}) must be greater than or equal the width of the second operand ({other.width}).")
        
        ext_other = other.__sext__(self.width)
        
        return icb(1 if self.repr != ext_other.repr else 0, self.width)
    

    def __sltu__(self, other: 'icb') -> 'icb':
        if self.width < other.width:
            raise ValueError(f"The width of the first operand ({self.width}) must be greater than or equal the width of the second operand ({other.width}).")
        
        ext_other = other.__sext__(self.width)
        
        return icb(1 if self.repr < ext_other.repr else 0, self.width)


    def __slt__(self, other: 'icb') -> 'icb':
        if self.width < other.width:
            raise ValueError(f"The width of the first operand ({self.width}) must be greater than or equal the width of the second operand ({other.width}).")
        
        ext_other = other.__sext__(self.width)

        self_sign_bit = self.repr >> (self.width - 1)
        other_sign_bit = ext_other.repr >> (ext_other.width - 1)
        if self_sign_bit > other_sign_bit:
            return icb(1, self.width)
        elif self_sign_bit < other_sign_bit:
            return icb(0, self.width)
        else:
            return self.__sltu__(ext_other)


    def __sleu__(self, other: 'icb') -> 'icb':
        if self.width < other.width:
            raise ValueError(f"The width of the first operand ({self.width}) must be greater than or equal the width of the second operand ({other.width}).")
        
        ext_other = other.__sext__(self.width)
        
        return icb(1 if self.repr <= ext_other.repr else 0, self.width)

    
    def __sle__(self, other: 'icb') -> 'icb':
        if self.width < other.width:
            raise ValueError(f"The width of the first operand ({self.width}) must be greater than or equal the width of the second operand ({other.width}).")
        
        ext_other = other.__sext__(self.width)

        self_sign_bit = self.repr >> (self.width - 1)
        other_sign_bit = ext_other.repr >> (ext_other.width - 1)
        if self_sign_bit > other_sign_bit:
            return icb(1, self.width)
        elif self_sign_bit < other_sign_bit:
            return icb(0, self.width)
        else:
            return self.__sleu__(ext_other)


    def __sgtu__(self, other: 'icb') -> 'icb':
        if self.width < other.width:
            raise ValueError(f"The width of the first operand ({self.width}) must be greater than or equal the width of the second operand ({other.width}).")
        
        ext_other = other.__sext__(self.width)
        
        return icb(1 if self.repr > ext_other.repr else 0, self.width)
    

    def __sgt__(self, other: 'icb') -> 'icb':
        if self.width < other.width:
            raise ValueError(f"The width of the first operand ({self.width}) must be greater than or equal the width of the second operand ({other.width}).")
        
        ext_other = other.__sext__(self.width)
        
        self_sign_bit = self.repr >> (self.width - 1)
        other_sign_bit = ext_other.repr >> (ext_other.width - 1)
        if self_sign_bit < other_sign_bit:
            icb(1, self.width)
        elif self_sign_bit > other_sign_bit:
            icb(0, self.width)
        else:
            return self.__sgtu__(ext_other)
        

    def __minu__(self, other: 'icb') -> 'icb':
        if self.width < other.width:
            raise ValueError(f"The width of the first operand ({self.width}) must be greater than or equal the width of the second operand ({other.width}).")
        
        ext_other = other.__sext__(self.width)

        return icb(self.repr if self.repr < ext_other.repr else ext_other.repr, self.width)
    

    def __min__(self, other: 'icb') -> 'icb':
        if self.width < other.width:
            raise ValueError(f"The width of the first operand ({self.width}) must be greater than or equal the width of the second operand ({other.width}).")
        
        ext_other = other.__sext__(self.width)

        self_sign_bit = self.repr >> (self.width - 1)
        other_sign_bit = ext_other.repr >> (ext_other.width - 1)
        if self_sign_bit > other_sign_bit:
            return icb(self.repr, self.width)
        elif self_sign_bit < other_sign_bit:
            return icb(ext_other.repr, ext_other.width)
        else:
            return self.__minu__(ext_other)
        

    def __maxu__(self, other: 'icb') -> 'icb':
        if self.width < other.width:
            raise ValueError(f"The width of the first operand ({self.width}) must be greater than or equal the width of the second operand ({other.width}).")
        
        ext_other = other.__sext__(self.width)

        return icb(self.repr if self.repr > ext_other.repr else ext_other.repr, self.width)
    

    def __max__(self, other: 'icb') -> 'icb':
        if self.width < other.width:
            raise ValueError(f"The width of the first operand ({self.width}) must be greater than or equal the width of the second operand ({other.width}).")
        
        ext_other = other.__sext__(self.width)

        self_sign_bit = self.repr >> (self.width - 1)
        other_sign_bit = ext_other.repr >> (ext_other.width - 1)
        if self_sign_bit < other_sign_bit:
            return icb(self.repr, self.width)
        elif self_sign_bit > other_sign_bit:
            return icb(ext_other.repr, ext_other.width)
        else:
            return self.__maxu__(ext_other)


    def to_bin(self) -> str:
        return f'0b{bin(self.repr)[2:].zfill(self.width)}'


    def to_hex(self) -> str:
        hex_width = (self.width + 3) // 4
        return f'0x{hex(self.repr)[2:].zfill(hex_width)}'


    @staticmethod
    def _sign_mag_to_two_comp(value: int, width: int) -> int:
        '''
        Convert a sign-magnitude integer to its two's complement representation.
        
        :param value: Sign-magnitude integer
        :param width: Bit width for the representation
        :return: 2's complement as integer
        '''
        
        if value >= 0:
            return value % (2 ** width)
        else:
            return ((~(-value) | (1 << (width - 1))) + 1) % (2 ** width)


    @staticmethod
    def parse_from_str(numeric_str: str) -> int:
        '''
        Parse a numeric string in sign-magnitude binary, hexadecimal, or decimal format and convert it to an integer.    
        
        :param numeric_str: A string representation of a number. 
                            It can be in binary format (e.g., '0b1010', '-0b0110'), 
                            hexadecimal format (e.g., '0x1A3F', '-0x1C2A'), 
                            or decimal format (e.g., '40', '-99').
        :return: The integer value corresponding to the input string.
        '''

        if not isinstance(numeric_str, str):
            raise ValueError(f"{numeric_str} is not a string")

        number_match = re.match(r'(?P<bin>-?0b[01]+)|(?P<hex>-?0x[0-9a-fA-F]+)|(?P<dec>-?\d+)', numeric_str)

        if not number_match:
            raise ValueError(f"'{numeric_str}' is not a numeric string")

        if number_match.group('bin'):
            return int(number_match.group('bin'), 2)
        elif number_match.group('hex'):
            return int(number_match.group('hex'), 16)
        elif number_match.group('dec'):
            return int(number_match.group('dec'))
        

    @staticmethod
    def get_bits(src: int, start: int, width: int) -> int:
        '''
        Extract a subset of bits from the source integer-coded-binary.

        :param src: The source integer from which bits are extracted.
        :type src: int
        :param start: The starting bit position (0-indexed, from the least significant bit).
        :type start: int
        :param width: The number of bits to extract.
        :type width: int
        :return: The extracted bits as an integer.
        :rtype: int
        :raises ValueError: If `width` is not positive or `start` is negative.

        **Examples**:

        .. code-block:: python

            # Example: Binary 1101101 (decimal 109)
            src = 0b1101101

            # Extract 3 bits starting from position 2
            result = get_bits(src, 2, 3)  # Output: 3 (Binary: 011)

            # Extract 4 bits starting from position 0
            result = get_bits(src, 0, 4)  # Output: 13 (Binary: 1101)

        '''
        if width <= 0:
            raise ValueError("Width must be a positive integer.")
        if start < 0:
            raise ValueError("Start position must be non-negative.")
        
        mask = (1 << width) - 1
        return (src >> start) & mask



ELEN = 32
VLEN = 128

V_ADDSUB = ['vadd', 'vsub', 'vrsub']
V_BITWISE = ['vand', 'vor', 'vxor']
V_SHIFT = ['vsll', 'vsrl', 'vsra']
V_COMPARE = ['vmseq', 'vmsne', 'vmsltu', 'vmslt', 'vmsleu', 'vmsle', 'vmsgtu', 'vmsgt']
V_MINMAX = ['vminu', 'vmin', 'vmaxu', 'vmax']
V_MERGEMV = ['vmerge', 'vmv']

V_ARITH = V_ADDSUB + V_BITWISE + V_SHIFT + V_COMPARE + V_MINMAX + V_MERGEMV
V_LOAD = ['vle', 'vlse', 'vluxei']
V_STORE = ['vse', 'vsse', 'vsuxei']

V_ALL = V_ARITH + V_LOAD + V_STORE


TYPE_WIDTHS = {# unit: bit
    '.byte': 8,
    '.half': 16,
    '.word': 32
}

DATA_TYPES = list(TYPE_WIDTHS.keys())

WIDTH_CODES = {
    8: 0b000,
    16: 0b101,
    32: 0b110
}


MOPS = {
    'vle': 0b00,
    'vluxei': 0b01,
    'vlse': 0b10,
    'vloxei': 0b11,

    'vse': 0b00,
    'vsuxei': 0b01,
    'vsse': 0b10,
    'vsoxei': 0b11,
}

VECTOR_INSTS = {
    'vadd': {'formats': ['vv', 'vx', 'vi'], 'group': 'v_arith/v_addsub', 'syntax': 'vadd.<format> <vd>, <vs2>, <vs1|rs1|imm>[, v0.t]', 'funct6': 0b000000},
    'vsub': {'formats': ['vv', 'vx'], 'group': 'v_arith/v_addsub', 'syntax': 'vsub.<format> <vd>, <vs2>, <vs1|rs1>[, v0.t]', 'funct6': 0b000010},
    'vrsub': {'formats': ['vx', 'vi'], 'group': 'v_arith/v_addsub', 'syntax': 'vrsub.<format> <vd>, <vs2>, <rs1|imm>[, v0.t]', 'funct6': 0b000011},
    'vand': {'formats': ['vv', 'vx', 'vi'], 'group': 'v_arith/v_bitwise', 'syntax': 'vand.<format> <vd>, <vs2>, <vs1|rs1|imm>[, v0.t]', 'funct6': 0b001001},
    'vor': {'formats': ['vv', 'vx', 'vi'], 'group': 'v_arith/v_bitwise', 'syntax': 'vor.<format> <vd>, <vs2>, <vs1|rs1|imm>[, v0.t]', 'funct6': 0b001010},
    'vxor': {'formats': ['vv', 'vx', 'vi'], 'group': 'v_arith/v_bitwise', 'syntax': 'vxor.<format> <vd>, <vs2>, <vs1|rs1|imm>[, v0.t]', 'funct6': 0b001011},
    'vsll': {'formats': ['vv', 'vx', 'vi'], 'group': 'v_arith/v_shift', 'syntax': 'vsll.<format> <vd>, <vs2>, <vs1|rs1|imm>[, v0.t]', 'funct6': 0b100101},
    'vsrl': {'formats': ['vv', 'vx', 'vi'], 'group': 'v_arith/v_shift', 'syntax': 'vsrl.<format> <vd>, <vs2>, <vs1|rs1|imm>[, v0.t]', 'funct6': 0b101000},
    'vsra': {'formats': ['vv', 'vx', 'vi'], 'group': 'v_arith/v_shift', 'syntax': 'vsra.<format> <vd>, <vs2>, <vs1|rs1|imm>[, v0.t]', 'funct6': 0b101001},
    'vmseq': {'formats': ['vv', 'vx', 'vi'], 'group': 'v_arith/v_compare', 'syntax': 'vmseq.<format> <vd>, <vs2>, <vs1|rs1|imm>[, v0.t]', 'funct6': 0b011000},
    'vmsne': {'formats': ['vv', 'vx', 'vi'], 'group': 'v_arith/v_compare', 'syntax': 'vmsne.<format> <vd>, <vs2>, <vs1|rs1|imm>[, v0.t]', 'funct6': 0b011001},
    'vmsltu': {'formats': ['vv', 'vx'], 'group': 'v_arith/v_compare', 'syntax': 'vmsltu.<format> <vd>, <vs2>, <vs1|rs1>[, v0.t]', 'funct6': 0b011010},
    'vmslt': {'formats': ['vv', 'vx'], 'group': 'v_arith/v_compare', 'syntax': 'vmslt.<format> <vd>, <vs2>, <vs1|rs1>[, v0.t]', 'funct6': 0b011011},
    'vmsleu': {'formats': ['vv', 'vx', 'vi'], 'group': 'v_arith/v_compare', 'syntax': 'vmsleu.<format> <vd>, <vs2>, <vs1|rs1|imm>[, v0.t]', 'funct6': 0b011100},
    'vmsle': {'formats': ['vv', 'vx', 'vi'], 'group': 'v_arith/v_compare', 'syntax': 'vmsle.<format> <vd>, <vs2>, <vs1|rs1|imm>[, v0.t]', 'funct6': 0b011101},
    'vmsgtu': {'formats': ['vx', 'vi'], 'group': 'v_arith/v_compare', 'syntax': 'vmsgtu.<format> <vd>, <vs2>, <rs1|imm>[, v0.t]', 'funct6': 0b011110},
    'vmsgt': {'formats': ['vx', 'vi'], 'group': 'v_arith/v_compare', 'syntax': 'vmsgt.<format> <vd>, <vs2>, <rs1|imm>[, v0.t]', 'funct6': 0b011111},
    'vminu': {'formats': ['vv', 'vx'], 'group': 'v_arith/v_minmax', 'syntax': 'vminu.<format> <vd>, <vs2>, <vs1|rs1>[, v0.t]', 'funct6': 0b000100},
    'vmin': {'formats': ['vv', 'vx'], 'group': 'v_arith/v_minmax', 'syntax': 'vmin.<format> <vd>, <vs2>, <vs1|rs1>[, v0.t]', 'funct6': 0b000101},
    'vmaxu': {'formats': ['vv', 'vx'], 'group': 'v_arith/v_minmax', 'syntax': 'vmaxu.<format> <vd>, <vs2>, <vs1|rs1>[, v0.t]', 'funct6': 0b000110},
    'vmax': {'formats': ['vv', 'vx'], 'group': 'v_arith/v_minmax', 'syntax': 'vmax.<format> <vd>, <vs2>, <vs1|rs1>[, v0.t]', 'funct6': 0b000111},
    'vmerge': {'formats': ['vvm', 'vxm', 'vim'], 'group': 'v_arith/v_mergemv', 'syntax': 'vmerge.<format> <vd>, <vs2>, <vs1|rs1|imm>, v0.t', 'funct6': 0b010111},
    'vmv': {'formats': ['v.v', 'v.x', 'v.i'], 'group': 'v_arith/v_mergemv', 'syntax': 'vmv.<format> <vd>, <vs1|rs1|imm>', 'funct6': 0b010111},
    'vle': {'widths': [8, 16, 32], 'group': 'v_load', 'syntax': 'vle<width>.v <vd>, (<rs1>)[, v0.t]'},
    'vlse': {'widths': [8, 16, 32], 'group': 'v_load', 'syntax': 'vlse<width>.v <vd>, (<rs1>), <rs2>[, v0.t]'},
    'vluxei': {'widths': [8, 16, 32], 'group': 'v_load', 'syntax': 'vluxei<width>.v <vd>, (<rs1>), <vs2>[, v0.t]'},
    'vse': {'widths': [8, 16, 32], 'group': 'v_store', 'syntax': 'vse<width>.v <vs3>, (<rs1>)[, v0.t]'},
    'vsse': {'widths': [8, 16, 32], 'group': 'v_store', 'syntax': 'vsse<width>.v <vs3>, (<rs1>), <rs2>[, v0.t]'},
    'vsuxei': {'widths': [8, 16, 32], 'group': 'v_store', 'syntax': 'vsuxei<width>.v <vs3>, (<rs1>), <vs2>[, v0.t]'},
}

SCALAR_INSTS = { # RV32I
    'lui': {'opcode': 0b0110111, 'group': 'u_type', 'syntax': 'lui <rd>, <imm20>'},
    'addi': {'opcode': 0b0010011, 'group': 'i_type', 'syntax': 'addi <rd>, <rs1>, <imm12>', 'funct3': 0b000}
}


SUPPORTED_INSTS = VECTOR_INSTS | SCALAR_INSTS # Merge


# Mapping
GROUP_OF_OPCODES = {
    0b0110111: 'u_type',
    0b0010011: 'i_type'
}

FUNCT6_TO_INST_MAP = {v['funct6']: k for k, v in VECTOR_INSTS.items() if 'funct6' in v}


from .icb import icb
from .supported_features import ELEN, VLEN

class vector:
    def __init__(self, vreg: int = 0, vect: list[icb] = None, elen: int = ELEN, vlen: int = VLEN) -> None:
        if vect != None:
            if len(vect) != vlen // elen:
                raise ValueError('Mismatch in size between source vector and configs (elen, vlen parameters)')
            
            self.elms = [elm.__sext__(elen) for elm in vect]
        else:
            self.elms = vector.vectorize(vreg, elen, vlen)
        self.elen = elen
        self.vlen = vlen
    
    
    def __vadd__(self, other: 'vector', masks: list[int] = None) -> 'vector':
        masks = masks or [1] * (self.vlen // self.elen)

        # Check mismatch
        if self.elen != other.elen or self.vlen != other.vlen:
            raise ValueError('Mismatch in size between two vector')

        result_vect = [
            self_elm.__add__(other_elm) if mask else icb(0, self.elen)
            for self_elm, other_elm, mask in zip(self.elms, other.elms, masks)
        ]

        return vector(vect=result_vect, elen=self.elen, vlen=self.vlen)


    def __vsub__(self, other: 'vector', masks: list[int] = None) -> 'vector':
        masks = masks or [1] * (self.vlen // self.elen)

        # Check mismatch
        if self.elen != other.elen or self.vlen != other.vlen:
            raise ValueError('Mismatch in size between two vector')

        result_vect = [
            self_elm.__sub__(other_elm) if mask else icb(0, self.elen)
            for self_elm, other_elm, mask in zip(self.elms, other.elms, masks)
        ]

        return vector(vect=result_vect, elen=self.elen, vlen=self.vlen)


    def __vrsub__(self, other: 'vector', masks: list[int] = None) -> 'vector':
        masks = masks or [1] * (self.vlen // self.elen)

        # Check mismatch
        if self.elen != other.elen or self.vlen != other.vlen:
            raise ValueError('Mismatch in size between two vector')

        result_vect = [
            other_elm.__sub__(self_elm) if mask else icb(0, self.elen)
            for self_elm, other_elm, mask in zip(self.elms, other.elms, masks)
        ]

        return vector(vect=result_vect, elen=self.elen, vlen=self.vlen)


    def __vand__(self, other: 'vector', masks: list[int] = None) -> 'vector':
        masks = masks or [1] * (self.vlen // self.elen)

        # Check mismatch
        if self.elen != other.elen or self.vlen != other.vlen:
            raise ValueError('Mismatch in size between two vector')

        result_vect = [
            self_elm.__and__(other_elm) if mask else icb(0, self.elen)
            for self_elm, other_elm, mask in zip(self.elms, other.elms, masks)
        ]

        return vector(vect=result_vect, elen=self.elen, vlen=self.vlen)


    def __vor__(self, other: 'vector', masks: list[int] = None) -> 'vector':
        masks = masks or [1] * (self.vlen // self.elen)

        # Check mismatch
        if self.elen != other.elen or self.vlen != other.vlen:
            raise ValueError('Mismatch in size between two vector')

        result_vect = [
            self_elm.__or__(other_elm) if mask else icb(0, self.elen)
            for self_elm, other_elm, mask in zip(self.elms, other.elms, masks)
        ]

        return vector(vect=result_vect, elen=self.elen, vlen=self.vlen)


    def __vxor__(self, other: 'vector', masks: list[int] = None) -> 'vector':
        masks = masks or [1] * (self.vlen // self.elen)

        # Check mismatch
        if self.elen != other.elen or self.vlen != other.vlen:
            raise ValueError('Mismatch in size between two vector')

        result_vect = [
            self_elm.__xor__(other_elm) if mask else icb(0, self.elen)
            for self_elm, other_elm, mask in zip(self.elms, other.elms, masks)
        ]

        return vector(vect=result_vect, elen=self.elen, vlen=self.vlen)


    def __vsll__(self, other: 'vector', masks: list[int] = None) -> 'vector':
        masks = masks or [1] * (self.vlen // self.elen)

        # Check mismatch
        if self.elen != other.elen or self.vlen != other.vlen:
            raise ValueError('Mismatch in size between two vector')

        result_vect = [
            self_elm.__sll__(other_elm) if mask else icb(0, self.elen)
            for self_elm, other_elm, mask in zip(self.elms, other.elms, masks)
        ]

        return vector(vect=result_vect, elen=self.elen, vlen=self.vlen)


    def __vsrl__(self, other: 'vector', masks: list[int] = None) -> 'vector':
        masks = masks or [1] * (self.vlen // self.elen)

        # Check mismatch
        if self.elen != other.elen or self.vlen != other.vlen:
            raise ValueError('Mismatch in size between two vector')

        result_vect = [
            self_elm.__srl__(other_elm) if mask else icb(0, self.elen)
            for self_elm, other_elm, mask in zip(self.elms, other.elms, masks)
        ]

        return vector(vect=result_vect, elen=self.elen, vlen=self.vlen)


    def __vsra__(self, other: 'vector', masks: list[int] = None) -> 'vector':
        masks = masks or [1] * (self.vlen // self.elen)

        # Check mismatch
        if self.elen != other.elen or self.vlen != other.vlen:
            raise ValueError('Mismatch in size between two vector')

        result_vect = [
            self_elm.__sra__(other_elm) if mask else icb(0, self.elen)
            for self_elm, other_elm, mask in zip(self.elms, other.elms, masks)
        ]

        return vector(vect=result_vect, elen=self.elen, vlen=self.vlen)


    def __vmseq__(self, other: 'vector', masks: list[int] = None) -> 'vector':
        masks = masks or [1] * (self.vlen // self.elen)

        # Check mismatch
        if self.elen != other.elen or self.vlen != other.vlen:
            raise ValueError('Mismatch in size between two vector')

        result_vect = [
            self_elm.__seq__(other_elm) if mask else icb(0, self.elen)
            for self_elm, other_elm, mask in zip(self.elms, other.elms, masks)
        ]

        return vector(vect=result_vect, elen=self.elen, vlen=self.vlen)


    def __vmsne__(self, other: 'vector', masks: list[int] = None) -> 'vector':
        masks = masks or [1] * (self.vlen // self.elen)

        # Check mismatch
        if self.elen != other.elen or self.vlen != other.vlen:
            raise ValueError('Mismatch in size between two vector')

        result_vect = [
            self_elm.__sne__(other_elm) if mask else icb(0, self.elen)
            for self_elm, other_elm, mask in zip(self.elms, other.elms, masks)
        ]

        return vector(vect=result_vect, elen=self.elen, vlen=self.vlen)


    def __vmsltu__(self, other: 'vector', masks: list[int] = None) -> 'vector':
        masks = masks or [1] * (self.vlen // self.elen)

        # Check mismatch
        if self.elen != other.elen or self.vlen != other.vlen:
            raise ValueError('Mismatch in size between two vector')

        result_vect = [
            self_elm.__sltu__(other_elm) if mask else icb(0, self.elen)
            for self_elm, other_elm, mask in zip(self.elms, other.elms, masks)
        ]

        return vector(vect=result_vect, elen=self.elen, vlen=self.vlen)


    def __vmslt__(self, other: 'vector', masks: list[int] = None) -> 'vector':
        masks = masks or [1] * (self.vlen // self.elen)

        # Check mismatch
        if self.elen != other.elen or self.vlen != other.vlen:
            raise ValueError('Mismatch in size between two vector')

        result_vect = [
            self_elm.__slt__(other_elm) if mask else icb(0, self.elen)
            for self_elm, other_elm, mask in zip(self.elms, other.elms, masks)
        ]

        return vector(vect=result_vect, elen=self.elen, vlen=self.vlen)


    def __vmsleu__(self, other: 'vector', masks: list[int] = None) -> 'vector':
        masks = masks or [1] * (self.vlen // self.elen)

        # Check mismatch
        if self.elen != other.elen or self.vlen != other.vlen:
            raise ValueError('Mismatch in size between two vector')

        result_vect = [
            self_elm.__sleu__(other_elm) if mask else icb(0, self.elen)
            for self_elm, other_elm, mask in zip(self.elms, other.elms, masks)
        ]

        return vector(vect=result_vect, elen=self.elen, vlen=self.vlen)


    def __vmsle__(self, other: 'vector', masks: list[int] = None) -> 'vector':
        masks = masks or [1] * (self.vlen // self.elen)

        # Check mismatch
        if self.elen != other.elen or self.vlen != other.vlen:
            raise ValueError('Mismatch in size between two vector')

        result_vect = [
            self_elm.__sle__(other_elm) if mask else icb(0, self.elen)
            for self_elm, other_elm, mask in zip(self.elms, other.elms, masks)
        ]

        return vector(vect=result_vect, elen=self.elen, vlen=self.vlen)


    def __vmsgtu__(self, other: 'vector', masks: list[int] = None) -> 'vector':
        masks = masks or [1] * (self.vlen // self.elen)

        # Check mismatch
        if self.elen != other.elen or self.vlen != other.vlen:
            raise ValueError('Mismatch in size between two vector')

        result_vect = [
            self_elm.__sgtu__(other_elm) if mask else icb(0, self.elen)
            for self_elm, other_elm, mask in zip(self.elms, other.elms, masks)
        ]

        return vector(vect=result_vect, elen=self.elen, vlen=self.vlen)


    def __vmsgt__(self, other: 'vector', masks: list[int] = None) -> 'vector':
        masks = masks or [1] * (self.vlen // self.elen)

        # Check mismatch
        if self.elen != other.elen or self.vlen != other.vlen:
            raise ValueError('Mismatch in size between two vector')

        result_vect = [
            self_elm.__sgt__(other_elm) if mask else icb(0, self.elen)
            for self_elm, other_elm, mask in zip(self.elms, other.elms, masks)
        ]

        return vector(vect=result_vect, elen=self.elen, vlen=self.vlen)


    def __vminu__(self, other: 'vector', masks: list[int] = None) -> 'vector':
        masks = masks or [1] * (self.vlen // self.elen)

        # Check mismatch
        if self.elen != other.elen or self.vlen != other.vlen:
            raise ValueError('Mismatch in size between two vector')

        result_vect = [
            self_elm.__minu__(other_elm) if mask else icb(0, self.elen)
            for self_elm, other_elm, mask in zip(self.elms, other.elms, masks)
        ]

        return vector(vect=result_vect, elen=self.elen, vlen=self.vlen)


    def __vmin__(self, other: 'vector', masks: list[int] = None) -> 'vector':
        masks = masks or [1] * (self.vlen // self.elen)

        # Check mismatch
        if self.elen != other.elen or self.vlen != other.vlen:
            raise ValueError('Mismatch in size between two vector')

        result_vect = [
            self_elm.__min__(other_elm) if mask else icb(0, self.elen)
            for self_elm, other_elm, mask in zip(self.elms, other.elms, masks)
        ]

        return vector(vect=result_vect, elen=self.elen, vlen=self.vlen)


    def __vmaxu__(self, other: 'vector', masks: list[int] = None) -> 'vector':
        masks = masks or [1] * (self.vlen // self.elen)

        # Check mismatch
        if self.elen != other.elen or self.vlen != other.vlen:
            raise ValueError('Mismatch in size between two vector')

        result_vect = [
            self_elm.__maxu__(other_elm) if mask else icb(0, self.elen)
            for self_elm, other_elm, mask in zip(self.elms, other.elms, masks)
        ]

        return vector(vect=result_vect, elen=self.elen, vlen=self.vlen)


    def __vmax__(self, other: 'vector', masks: list[int] = None) -> 'vector':
        masks = masks or [1] * (self.vlen // self.elen)

        # Check mismatch
        if self.elen != other.elen or self.vlen != other.vlen:
            raise ValueError('Mismatch in size between two vector')

        result_vect = [
            self_elm.__max__(other_elm) if mask else icb(0, self.elen)
            for self_elm, other_elm, mask in zip(self.elms, other.elms, masks)
        ]

        return vector(vect=result_vect, elen=self.elen, vlen=self.vlen)


    def __vmerge__(self, other: 'vector', masks: list[int] = None) -> 'vector':
        masks = masks or [1] * (self.vlen // self.elen)

        # Check mismatch
        if self.elen != other.elen or self.vlen != other.vlen:
            raise ValueError('Mismatch in size between two vector')

        result_vect = [
            other_elm if mask else self_elm
            for self_elm, other_elm, mask in zip(self.elms, other.elms, masks)
        ]

        return vector(vect=result_vect, elen=self.elen, vlen=self.vlen)
    

    def to_register(self) -> int:
        '''
        Convert the vector as list of 'icb' type to a vector register as integer
        '''
        vreg = 0
        for index, elm in enumerate(self.elms):
            vreg |= (elm.repr << (self.elen * index))
        
        return vreg
    
    
    def get_element(self, index: int) -> icb:
        if index < 0 or index > self.vlen // self.elen:
            raise ValueError('Invalid index')
        
        return self.elms[index]
    

    @staticmethod
    def vectorize(vreg: int, elen: int, vlen: int) -> list[icb]:
        '''
        Convert a vector register as integer (it codes binary - icb) to list of 'icb' type
        '''
        vect = []
        for i in range(vlen // elen):
            elm = icb.get_bits(vreg, start=(i * elen), width=elen)
            vect.append(icb(elm, elen))
        return vect


from random import randint, choice
from ..common.helpers import repeated_random
from ..common.supported_features import V_LOAD, V_STORE, VECTOR_INSTS, SCALAR_INSTS, TYPE_WIDTHS
from .ga import run_evolution


def generate_operator_sequence(configs):
    '''
    Receive the configs and run evoluation from genetic algorithm
    to generate operator sequence.
    '''
    operator_seq, _ = run_evolution(configs)
    return operator_seq


def add_operands(operator_sequence, configs):
    '''
    Receive a operator sequence and add operands
    to get the completed instruction sequence.
    '''
    
    used_vd_regs = set()
    used_x_regs = set()

    completed_inst_sequence = []

    for operator in operator_sequence:
        inst = {
            'operator': operator,
            'vm': True if configs['vector_masking'] else choice([True, False])
            # vm = 1 --> asm_vm = nothing --> unmasked
            # vm = 0 --> asm_vm = v0.t    --> vector result, only where v0[i][0] = 1
        }

        group = VECTOR_INSTS[operator]['group']
        if group.startswith('v_arith'):
            # Randomize instruction format
            format = choice(VECTOR_INSTS[operator]['formats'])
            inst['format'] = format
            
            # Randomize vs2 register
            inst['vs2'] = repeated_random(0, 31, used_vd_regs, configs['v_reg_reuse_rate'])

            if format in ['vv', 'v.v', 'vvm']:
                inst['vs1'] = repeated_random(0, 31, used_vd_regs, configs['v_reg_reuse_rate'])
            elif format in ['vx', 'v.x', 'vxm']:
                inst['rs1'] = repeated_random(0, 31, used_x_regs, configs['x_reg_reuse_rate'])
                used_x_regs.add(inst['rs1'])
            elif format in ['vi', 'v.i', 'vim']:
                inst['imm'] = randint(0b00000, 0b11111)

            inst['vd'] = randint(0, 31)
            used_vd_regs.add(inst['vd'])
        
        else: # V_LOAD or V_STORE
            inst['width'] = choice(VECTOR_INSTS[operator]['widths'])
            inst['rs1'] = repeated_random(1, 31, used_x_regs, configs['x_reg_reuse_rate'])
            used_x_regs.add(inst['rs1'])
          
            if operator == 'vlse' or operator == 'vsse':
                inst['rs2'] = randint(0, 31)
                used_x_regs.add(inst['rs2'])
            elif operator == 'vluxei' or operator == 'vsuxei':
                inst['vs2'] = repeated_random(0, 31, used_vd_regs, configs['v_reg_reuse_rate'])

            if operator in V_LOAD:
                inst['vd'] = randint(0, 31)
                used_vd_regs.add(inst['vd'])
            elif operator in V_STORE:
                inst['vs3'] = repeated_random(0, 31, used_vd_regs, configs['v_reg_reuse_rate'])

        completed_inst_sequence.append(inst)

    return completed_inst_sequence


def insert_vm_change_placeholders(inst_sequence: list, configs):
    num_of_insts = configs['num_of_insts']
    num_of_vms = 0 if (configs['vector_masking']) else \
                 int(num_of_insts * configs['vector_masking_change_rate'])

    if num_of_vms == 0:
        return inst_sequence
    
    strip = num_of_insts // num_of_vms

    inst_sequence.insert(0, {
        'operator': 'placeholder.chgvm',
        'vm_label': f'__vm0__'
    })
    start_pos = 1
    for i in range(1, num_of_vms + 1):
        insert_pos = randint(start_pos, start_pos + strip - 1)

        inst_sequence.insert(insert_pos + 1, {
            'operator': 'placeholder.chgvm',
            'vm_label': f'__vm{i}__'
        })

        start_pos = strip * i + 1 + 1

    return inst_sequence


def replace_vm_change_placeholders(inst_sequence: list):
    updated_sequence = []

    for inst in inst_sequence:
        if inst['operator'] == 'placeholder.chgvm':
            vm_label = inst['vm_label']

            imm20 = f"%hi({vm_label})"
            imm12 = f"%lo({vm_label})"

            updated_sequence.extend([
                {'operator': 'lui', 'rd': 1, 'imm20': imm20},
                {'operator': 'addi', 'rd': 1, 'rs1': 1, 'imm12': imm12},
                {'operator': 'vle', 'width': 8, 'vd': 0, 'rs1': 1, 'vm': True}
            ])
        else:
            updated_sequence.append(inst)
    
    return updated_sequence


def display_inst(inst):
    print(inst)
    operator = inst['operator']
    if operator in VECTOR_INSTS:
        group = VECTOR_INSTS[operator]['group']
        if group.startswith('v_load') or group.startswith('v_store'):
            assembly = f"{operator}{inst['width']}.v "

            if operator in V_LOAD:
                assembly += f"v{inst['vd']}, (x{inst['rs1']})"
            elif operator in V_STORE:
                assembly += f"v{inst['vs3']}, (x{inst['rs1']})"
            
            if operator in ['vlse', 'vsse']:
                assembly += f", x{inst['rs2']}"
            elif operator in ['vluxei', 'vsuxei']:
                assembly += f", v{inst['vs2']}"

            if not (inst['vm']):
                assembly += ', v0.t'

            return assembly
    
        else:
            if operator == 'vmerge':
                if inst['format'] == 'vvm':
                    return f"vmerge.vvm v{inst['vd']}, v{inst['vs2']}, v{inst['vs1']}, v0.t"
                elif inst['format'] == 'vxm':
                    return f"vmerge.vxm v{inst['vd']}, v{inst['vs2']}, x{inst['rs1']}, v0.t"
                elif inst['format'] == 'vim':
                    return f"vmerge.vim v{inst['vd']}, v{inst['vs2']}, 0x{inst['imm']:x}, v0.t"

            elif operator == 'vmv':
                if inst['format'] == 'v.v':
                    return f"vmv.v.v v{inst['vd']}, v{inst['vs1']}"
                elif inst['format'] == 'v.x':
                    return f"vmv.v.x v{inst['vd']}, x{inst['rs1']}"
                elif inst['format'] == 'v.i':
                    return f"vmv.v.i v{inst['vd']}, 0x{inst['imm']:x}"
            
            else:
                assembly = f"{operator}.{inst['format']} v{inst['vd']}, v{inst['vs2']}, "
                if inst['format'] == 'vv':
                    assembly += f"v{inst['vs1']}"
                elif inst['format'] == 'vx':
                    assembly += f"x{inst['rs1']}"
                elif inst['format'] == 'vi':
                    assembly += f"0x{inst['imm']:x}"

                if not (inst['vm']):
                    assembly += ', v0.t'

                return assembly
        
    elif operator in SCALAR_INSTS:
        group = SCALAR_INSTS[operator]['group']
        if group.startswith('i_type'):
            imm12 = inst['imm12']
            return f"{operator} x{inst['rd']}, x{inst['rs1']}, {imm12 if type(imm12) == str else hex(imm12)}"
        elif group.startswith('u_type'):
            imm20 = inst['imm20']
            return f"{operator} x{inst['rd']}, {imm20 if type(imm20) == str else hex(imm20)}"
        
    else:
        raise ValueError(f"Unsupported operator: '{operator}'")


def display_data_def(label, data_def):
    return f"{label}: {data_def['type']} {', '.join([hex(item) for item in data_def['items']])}"
    

def display(
    data_section,
    init_text_seq,
    main_text_seq,
):
    TAB = 4 * ' '

    asm_data_section = ''
    for label, data_def in data_section.items():
        asm_data_section += display_data_def(label, data_def) + '\n'

    asm_init_text = ''
    asm_init_sequence = [display_inst(inst) for inst in init_text_seq]
    for inst in asm_init_sequence:
        asm_init_text +=f'{TAB}{inst}\n'


    asm_main_text = ''
    asm_main_sequence = [display_inst(inst) for inst in main_text_seq]
    for inst in asm_main_sequence:
        asm_main_text += f'{TAB}{inst}\n'

    return '.data\n' + \
        asm_data_section + \
        '\n' + \
        '.text\n' + \
        'init:\n' + \
        asm_init_text + \
        'main:\n' + \
        asm_main_text




import json
from ..common.helpers import ask_user



CONFIG_SCHEMAS = {
    'num_of_tests': {
        'type': int,
        'required': True,
        'min': 1,
    },
    'test_prefix': {
        'type': str,
        'default': 'test',
    },
    'num_of_insts': {
        'type': int,
        'required': True,
        'min': 10,
    },
    'v_arith': {
        'type': [int, float],
        'default': 14,
    },
    'v_load': {
        'type': [int, float],
        'default': 3,
    },
    'v_store': {
        'type': [int, float],
        'default': 3,
    },
    'v_reg_reuse_rate': {
        'type': float,
        'default': 0.3,
        'min': 0.0,
        'max': 1.0,
        'typical_min': 0.1,
        'typical_max': 0.4,
    },
    'x_reg_reuse_rate': {
        'type': float,
        'default': 0.1,
        'min': 0.0,
        'max': 1.0,
        'typical_min': 0.05,
        'typical_max': 0.3,
    },
    "data_type_weights": {
        'type': list,
        'length': 3,
        'item_type': [int, float],
        'default': [1, 1, 1] # [.byte, .half, .word]
    },
    "data_variant_weights": {
        'type': list,
        'length': 4,
        'item_type': [int, float],
        'default': [1, 3, 3, 3] # [zero, neg, pos, big_num]
    },
    'vector_masking': {
        'type': bool,
        'default': False
    },
    "vector_masking_change_rate": { # this schema must be defined after 'vector_masking'
        'type': float,
        'default': 0.2,
        'min': 0.0,
        'max': 1.0,
        'typical_min': 0.05,
        'typical_max': 0.3,
    },
    'dmem_size': { # TODO: check align, multiples of 4
        'type': int,
        'required': True,
        'min': 4 * (2 ** 10), # 4096 bytes = 1024 words  
    },
    'dmem_access_range_start': { # TODO: check align, multiples of 4
        'type': int,
        'default': 0,
    },
    'dmem_access_range_end': { # TODO: check align, multiples of 4
        'type': int,
        'default': '$dmem_size',
    },
    'rtg': {
        'mutate_rate': {
            'type': float,
            'default': 0.1,
            'min': 0.0,
            'max': 1.0,
            'typical_min': 0.05,
            'typical_max': 0.2,
        },
        'population_size': {
            'type': int,
            'default': 10,
            'min': 5,
            'typical_min': 5,
            'typical_max': 20,    
        }
    },
}

def _validate_configs(configs, schemas):
    output_configs = {}

    for key, schema in schemas.items():
        if key not in configs:
            if key =='rtg': # sub-schemas
                output_configs['rtg'] = _validate_configs({}, schema)
                continue

            is_required = schema.get('required', False)
            if is_required:
                print(f'[Config Reading][Error] \'{key}\' is required.')
                exit(-1)
            else:
                default_value = schema['default']
                if type(default_value) == str and default_value[0] == '$':
                    output_configs[key] = output_configs[default_value[1:]]
                    print(f'[Config Reading][Info] \'{key}\' is not defined, take default: \'{output_configs[default_value[1:]]}\'.')
                else:
                    output_configs[key] = default_value
                    print(f'[Config Reading][Info] \'{key}\' is not defined, take default: \'{default_value}\'.')
                continue
        
        config = configs[key]

        # sub-schemas
        if key == 'rtg':
            if type(config) != dict:
                print(f'[Config Reading][Error] \'{key}\' is must be a sub-config.')
                exit(-1)

            output_configs['rtg'] = _validate_configs(config, schema)
            configs.pop(key)
            continue
        
        # check dependent configs
        if key == 'vector_masking_change_rate' and output_configs['vector_masking']:
            print(f'[Config Reading][Info] \'{key}\' is useless in case \'vector_masking\' is True (unmasked).')
            configs.pop(key)
            continue

        # check type
        expected_type = schema['type']
        current_type = type(config)
        if (type(expected_type) == list): # check multi type
            if current_type not in expected_type:
                print(f"[Config Reading][Error] '{key}' must be {' or '.join([t.__name__ for t in expected_type])}, but it is {current_type.__name__}.")
                exit(-1)
        elif expected_type == list: # check list
            expected_length = schema['length']
            current_length = len(config)
            if current_length != expected_length: # check length
                print(f"[Config Reading][Error] '{key}' must be list of {expected_length} items, but it contains {current_length}.")
                exit(-1)
            expected_item_type = schema['item_type']
            if type(expected_item_type) == list:
                if any(type(item) not in expected_item_type for item in config): # check multi type of items
                    print(f"[Config Reading][Error] '{key}' must be list of {'s or '.join([t.__name__ for t in expected_item_type])}s.")
                    exit(-1)
            else:
                if any(type(item) != expected_item_type for item in config): # check single type of items
                    print(f"[Config Reading][Error] '{key}' must be list of {expected_item_type}s.")
                    exit(-1)
        else:
            if current_type != expected_type: # check single type
                print(f'[Config Reading][Error] \'{key}\' must be {expected_type.__name__}, but it is {current_type.__name__}.')
                exit(-1)

        # check range
        min = schema.get('min', None)
        max = schema.get('max', None)
        if min is not None and config < min:
            print(f'[Config Reading][Error] \'{key}\' must be greater than \'{min}\'.')
            exit(-1)

        if max is not None and config > max:
            print(f'[Config Reading][Error] \'{key}\' must be less than \'{max}\'.')
            exit(-1)

        # check typical range
        typical_min = schema.get('typical_min', None)
        typical_max = schema.get('typical_max', None)
        if typical_min is not None and config < typical_min:
            print(f'[Config Reading][Warning] \'{key}\' should be greater than \'{typical_min}\'.')
            if not ask_user('Do you want continue?'):
                exit(1)

        if typical_max is not None and config > typical_max:
            print(f'[Config Reading][Warning] \'{key}\' should be less than \'{typical_max}\'.')
            if not ask_user('Do you want continue?'):
                exit(1)

        # add to output configs and remove from input configs
        output_configs[key] = config
        configs.pop(key)

    # check invalid for unsupported configs
    for key, config in configs.items():
        print(f'[Config Reading][Warning] \'{key}\' is invalid or unsuppported and will be ignored.')
        if not ask_user('Do you want continue?'):
            exit(1)

    return output_configs


def parse_config_file(config_file):
    f = open(config_file, 'r')
    configs = json.load(f)
    f.close()

    good_configs = _validate_configs(configs, CONFIG_SCHEMAS)

    return good_configs


from random import choice, choices, randint, random
from ..common.supported_features import V_ALL, V_ARITH, V_LOAD, V_STORE

def _init_population(population_size, chromosome_len, gene_pool):
    population = []
    for _ in range(population_size):
        chromosome = choices(gene_pool, k=chromosome_len)
        population.append(chromosome)

    return population


def _evaluate_fitness(chromosome, configs):
    counters = {
        'v_load': 0,
        'v_store': 0,
        'v_arith': 0
    }

    for gene in chromosome:
        if gene in V_ARITH:
            counters['v_arith'] += 1
        elif gene in V_LOAD:
            counters['v_load'] += 1
        elif gene in V_STORE:
            counters['v_store'] += 1

    difference = sum(abs(counters[key] - configs[key]) for key in counters)

    return difference


def _crossover_1_point(parent_1, parent_2):
    p = randint(1, len(parent_1) - 1)
    offspring_1 = parent_1[:p] + parent_2[p:]
    offspring_2 = parent_2[:p] + parent_1[p:]

    return offspring_1, offspring_2


def _let_it_be_mutated(MUTATE_RATE):
    r = random()
    return (r <= MUTATE_RATE)


def _mutate(chromosome, gene_pool):
    mutate_point = randint(0, len(chromosome) - 1)
    while True:
        newGene = choice(gene_pool)
        if newGene != chromosome[mutate_point]:
            chromosome[mutate_point] = newGene
            break

    return chromosome

def run_evolution(configs):
    POPULATION_SIZE = configs['rtg']['population_size']
    CHROMOSOME_LEN = configs['num_of_insts']
    generation_counter = 0
    population = _init_population(
        population_size=POPULATION_SIZE,
        chromosome_len=CHROMOSOME_LEN,
        gene_pool=V_ALL
    )
    
    while (True):
        population.sort(key=lambda chromosome: _evaluate_fitness(chromosome, configs))
        
        if _evaluate_fitness(population[0], configs) == 0:
            break

        new_generation = []

        # Perform Elitism, that mean 10% of fittest population 
        # goes to the next generation 
        s = round((10 * configs['rtg']['population_size'] ) / 100)
        if s < 1: s = 1
        new_generation.extend(population[:s]) 
        
        rest = configs['rtg']['population_size'] - s
        half = round((50 * configs['rtg']['population_size']) / 100) 
        for _ in range(rest): 
            parent_1 = choice(population[:half+1])
            while True:
                parent_2 = choice(population[:half+1])
                if parent_2 != parent_1:
                    break

            offspring_1, offspring_2 = _crossover_1_point(parent_1, parent_2)

            if _let_it_be_mutated(configs['rtg']['mutate_rate']):
                offspring_1 = _mutate(offspring_1, V_ALL)
            if _let_it_be_mutated(configs['rtg']['mutate_rate']):
                offspring_2 = _mutate(offspring_2, V_ALL)
            
            new_generation.append(offspring_1) 
            new_generation.append(offspring_2) 
  
        population = new_generation 
  
        generation_counter += 1

    return population[0], generation_counter
    



import setuptools

with open("README.md", "r", encoding = "utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name = "rvvsuite",
    version = "0.0.5",
    author = "Nguyen Binh Khiem",
    author_email = "khiemnb153@gmail.com",
    description = "A set of tools for developing RISC-V Vector IP includes Random Test Generator (RTG), Assembler, and Simulator",
    long_description = long_description,
    long_description_content_type = "text/markdown",
    url='https://github.com/khiemnb153/rvvsuite',
    classifiers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir = {"": "src"},
    packages = setuptools.find_packages(where="src"),
    python_requires = ">=3.13"
)